# -*- coding: utf-8 -*-
# Based heavily on the implementation of pint's numpy array function wrapping

from __future__ import annotations

try:
    import numpy as np
except ImportError:
    np = None
try:
    import jax
    import jax.numpy as jnp

except ImportError:
    jax = None
    jnp = None

from .util import has_length, is_iterable, ndarray_to_scalar

HANDLED_UFUNCS = {}
HANDLED_FUNCTIONS = {}


def _is_uncertainty(obj):
    """Test for _nom and _err attrs.

    This is done in place of isinstance(Uncertainty, arg), which would cause a circular import.

    Parameters
    ----------
    obj : Object


    Returns
    -------
    bool
    """
    return hasattr(obj, "_nom") and hasattr(obj, "_err")


def _is_sequence_with_uncertainty_elements(obj):
    """Test for sequences of uncertainties.

    Parameters
    ----------
    obj : object


    Returns
    -------
    True if obj is a sequence and at least one element is an Uncertainty; False otherwise
    """
    return (
        is_iterable(obj)
        and has_length(obj)
        and not isinstance(obj, str)
        and any(_is_uncertainty(item) for item in obj)
    )


def convert_arg(arg, attr: str = None):
    """Convert uncertainties and sequences of uncertainties to nominal values or errors

    This function has a different behavior if the nominal value or errors are requested.
    If the nominal value is requested, the outputs are:
        copies if arg is not an uncertainty (or sequence of)
        nominal values (or sequence of) if arg is an uncertainty (or sequence of)
    If the errors are requested, the outputs are:
        None if arg is not an uncertainty (or sequence of)
        errors (or sequence of) if arg is an uncertainty (or sequence of)
    """
    if _is_uncertainty(arg):
        return getattr(arg, attr)
    elif _is_sequence_with_uncertainty_elements(arg):
        if attr != "_nom":
            return None
        else:
            return [convert_arg(item, attr) for item in arg]
    else:
        if attr != "_nom":
            return None
        else:
            return arg


def classify_and_split_args_and_kwargs(*args, **kwargs):
    """Checks the args and kwargs to see if they contain uncertanty info, and prepares them for use by JAX.

    JAX does not support differentiating with respect to kwargs so uncertainty info there is just discarded.
    Returns:
        uncert_argnums: list of int
            A list of the positional arguments with respect to which derivatives need to be taken
        uncert_arg_nom:
            A list of arguments to the function, without errors
        uncert_arg_err:
            A list of errors for the positional arguments which will have derivatives
        uncert_kwarg_nom:
            A dict of keyword args that will be passed to the function
        uncert_instance:
            Returns an instance of an Uncertainty object for the class constructor
    """

    uncert_argnums = tuple(
        idx
        for idx, arg in enumerate(args)
        if convert_arg(arg, "_nom") is not None
    )
    uncert_arg_nom = tuple(convert_arg(arg, "_nom") for arg in args)
    uncert_arg_err = []
    for aidx, arg in enumerate(args):
        carg = convert_arg(arg, "_err")
        if convert_arg(arg, "_err") is not None:
            uncert_arg_err.append(carg)
        else:
            uncert_arg_err.append(jnp.zeros_like(uncert_arg_nom[aidx]))
    uncert_arg_err = tuple(uncert_arg_err)
    uncert_kwarg_nom = {
        key: convert_arg(arg, "_nom") for key, arg in kwargs.items()
    }
    return uncert_argnums, uncert_arg_nom, uncert_arg_err, uncert_kwarg_nom


def implements(numpy_func_string, func_type):
    """Register an __array_function__/__array_ufunc__ implementation for Uncertainty
    objects.

    """

    def decorator(func):
        if func_type == "function":
            HANDLED_FUNCTIONS[numpy_func_string] = func
        elif func_type == "ufunc":
            HANDLED_UFUNCS[numpy_func_string] = func
        else:
            raise ValueError("Invalid func_type {}".format(func_type))
        return func

    return decorator


def get_func_from_package(func_str, namespace):
    # Handle functions in submodules
    func_str_split = func_str.split(".")
    func = getattr(namespace, func_str_split[0], None)
    # If the function is not available, do not attempt to implement it
    if func is None:
        return
    for func_str_piece in func_str_split[1:]:
        func = getattr(func, func_str_piece)

    return func


def elementwise_grad(g):
    def wrapped(*args, **kwargs):
        y, g_vjp = jax.vjp(lambda *a: g(*a, **kwargs), *args)
        x_bar = g_vjp(np.ones_like(y))
        return x_bar

    return wrapped


def get_mappable_dims(*args):
    # Check that all the args have the same dimension
    assert all([a.ndim == args[0].ndim for a in args])
    # Check that the size of each dimension is either the same as the maximum, or 1
    mappable = [None for a in args]
    max_dim_sizes = []
    for i, dim in enumerate(range(args[0].ndim)):
        sz = [a.shape[dim] for a in args]
        max_sz = max(sz)
        max_dim_sizes.append(max_sz)
        assert all([s == max_sz or s == 1 for s in sz])
    for i, a in enumerate(args):
        map_axes = []
        for j, dim in enumerate(range(args[0].ndim)):
            if a.shape[dim] == max_dim_sizes[j]:
                map_axes.append(j)
        if len(map_axes) > 1:
            mappable[i] = tuple(map_axes)
        elif len(map_axes) == 1:
            mappable[i] = map_axes[0]
    return mappable, max_dim_sizes


def implement_func(
    func_type,
    func_str,
    implement_mode,
    grad_argnum_override=None,
    selection_operator=None,
    output_rank=0,
):
    """Add default-behavior NumPy function/ufunc to the handled list.

    Parameters
    ----------
    func_type : str
        "function" for NumPy functions, "ufunc" for NumPy ufuncs
    func_str : str
        String representing the name of the NumPy function/ufunc to add
    implement_mode: str
        Instructs on the implementation type
    grad_argnum_override: list of int
        Positions of arguments that should be differentiated, if necessary to enforce
    selection_operator: func
        An operator that provides selection indices that correspond to the action of the function that needs to be implemented
        E.g. np.argmax for np.max
    output_rank: int
        The rank of the output. If it's greater than rank 0 and derivatives are needed, jacfwd needs to be used instead of grad.
    """
    # If Jax+NumPy is not available, do not attempt implement that which does not exist
    if jnp is None:
        return

    func = get_func_from_package(func_str, jnp)
    # Skip the JAX overhead if you dont need gradient info
    func_np = get_func_from_package(func_str, np)

    @implements(func_str, func_type)
    def implementation(*args, **kwargs):
        from auto_uncertainties import Uncertainty

        (
            uncert_argnums,
            uncert_arg_nom,
            uncert_arg_err,
            uncert_kwarg_nom,
        ) = classify_and_split_args_and_kwargs(*args, **kwargs)

        # Determine result through base numpy function on stripped arguments
        if implement_mode == "same_shape":
            bcast_args_nom = np.broadcast_arrays(*uncert_arg_nom)
            bcast_args_err = np.broadcast_arrays(*uncert_arg_err)
            value = func_np(*bcast_args_nom, **uncert_kwarg_nom)
            grads = elementwise_grad(func)(*bcast_args_nom, **uncert_kwarg_nom)
            error_dot_grad_sqr = [
                (e * g) ** 2 for e, g in zip(bcast_args_err, grads)
            ]
            error = np.sum(error_dot_grad_sqr, axis=0) ** 0.5
            return Uncertainty(value, error)
        #            return uncert_instance.__class__(val, err)
        elif implement_mode == "same_shape_bool":
            return func_np(*uncert_arg_nom, **uncert_kwarg_nom)
        elif implement_mode == "nograd":
            return func_np(*uncert_arg_nom, **uncert_kwarg_nom)
        elif implement_mode == "selection_operator":
            return func_np(*uncert_arg_nom, **uncert_kwarg_nom)
        elif implement_mode == "selection":
            sel_func_np = get_func_from_package(selection_operator, np)
            axis = uncert_kwarg_nom.pop("axis", None)
            if axis is None:
                idx = sel_func_np(*uncert_arg_nom, **uncert_kwarg_nom)
                return np.ravel(uncert_arg_nom[0])[idx]
            else:
                idxs = np.expand_dims(
                    sel_func_np(
                        *uncert_arg_nom, axis=axis, **uncert_kwarg_nom
                    ),
                    axis=axis,
                )
                return np.take_along_axis(uncert_arg_nom[0], idxs, axis=axis)
        elif implement_mode in ["apply_to_both"]:
            val = func_np(*uncert_arg_nom, **uncert_kwarg_nom)
            err = np.abs(func_np(*uncert_arg_err, **uncert_kwarg_nom))
            return Uncertainty(val, err)
        elif implement_mode in ["reduction_binary", "reduction_unary"]:
            axis = uncert_kwarg_nom.get("axis", None)

            bcast_args_nom = np.broadcast_arrays(*uncert_arg_nom)
            bcast_args_err = np.broadcast_arrays(*uncert_arg_err)

            val = func_np(*bcast_args_nom, **uncert_kwarg_nom)
            grads = elementwise_grad(func)(*bcast_args_nom, **uncert_kwarg_nom)
            if axis is not None:
                axis = tuple(axis)
                error_dot_grad_sqr = [
                    np.sum((e * g) ** 2, axis=axis)
                    for e, g in zip(bcast_args_err, grads)
                ]
            else:
                error_dot_grad_sqr = [
                    np.sum((e * g) ** 2) for e, g in zip(bcast_args_err, grads)
                ]

            err = np.sum(error_dot_grad_sqr, axis=0) ** 0.5
            return Uncertainty(ndarray_to_scalar(val), ndarray_to_scalar(err))


# Returns a bool array of the same shape (i.e. elementwise conditionals)
unary_bcast_same_shape_bool_ufuncs = [
    "isnan",
    "isinf",
    "isfinite",
    "signbit",
]
binary_bcast_same_shape_bool_ufuncs = [
    "equal",
    "greater",
    "greater_equal",
    "less",
    "less_equal",
    "not_equal",
]
bcast_same_shape_bool_ufuncs = (
    binary_bcast_same_shape_bool_ufuncs + unary_bcast_same_shape_bool_ufuncs
)
for ufunc in bcast_same_shape_bool_ufuncs:
    implement_func("ufunc", ufunc, implement_mode="same_shape_bool")

# Applies the ufunc to the value and discards the error
bcast_nograd_ufuncs = [
    "sign",
    "floor_divide",
    "fmod",
    "mod",
    "remainder",
    "copysign",
    "nextafter",
    "trunc",
    "spacing",
]
for ufunc in bcast_nograd_ufuncs:
    implement_func("ufunc", ufunc, implement_mode="nograd")
# Returns an Uncertainty array of the same shape using the gradient (i.e. elementwise math operations)
unary_bcast_same_shape_ufuncs = [
    "sqrt",
    "cbrt",
    "square",
    "reciprocal",
    "arccos",
    "arcsin",
    "arctan",
    "arccosh",
    "arcsinh",
    "arctanh",
    "exp",
    "expm1",
    "exp2",
    "log",
    "log10",
    "log1p",
    "log2",
    "sin",
    "cos",
    "tan",
    "sinh",
    "cosh",
    "tanh",
    "radians",
    "degrees",
    "deg2rad",
    "rad2deg",
    "absolute",
]
binary_bcast_same_shape_ufuncs = [
    "add",
    "subtract",
    "multiply",
    "divide",
    "true_divide",
    "logaddexp",
    "logaddexp2",
    "divide",
    "arctan2",
    "hypot",
]
bcast_same_shape_ufuncs = (
    binary_bcast_same_shape_ufuncs + unary_bcast_same_shape_ufuncs
)
for ufunc in bcast_same_shape_ufuncs:
    implement_func("ufunc", ufunc, implement_mode="same_shape")

# Returns the indices of the Uncertainty array by some criteria
bcast_selection_operator_funcs = ["argmax", "argmin"]

for ufunc in bcast_selection_operator_funcs:
    implement_func("function", ufunc, implement_mode="selection_operator")

# Selects a sub-section of or reshapes the Uncertainty array by some criteria
bcast_selection_funcs = {
    "max": "argmax",
    "min": "argmin",
    "amax": "argmax",
    "amin": "argmin",
}

for ufunc, sel_op in bcast_selection_funcs.items():
    implement_func(
        "function",
        ufunc,
        implement_mode="selection",
        selection_operator=sel_op,
    )

# Applies ufunc or func to both the value and error
bcast_apply_to_both_funcs = [
    "compress",
    "diagonal",
    "ravel",
    "repeat",
    "reshape",
    "squeeze",
    "swapaxes",
    "take",
    "transpose",
]
bcast_apply_to_both_ufuncs = [
    "conj",
    "conjugate",
    "negative",
    "positive",
    "fabs",
    "round",
    "ceil",
    "floor",
    "rint",
]
for ufunc in bcast_apply_to_both_ufuncs:
    implement_func("ufunc", ufunc, implement_mode="apply_to_both")
for ufunc in bcast_apply_to_both_funcs:
    implement_func("function", ufunc, implement_mode="apply_to_both")

# Applies a reduction
implement_func("function", "trapz", implement_mode="reduction_binary")

bcast_reduction_unary = ["std", "sum", "var", "mean", "ptp", "median"]
for ufunc in bcast_reduction_unary:
    implement_func("function", ufunc, implement_mode="reduction_unary")


@implements("round", "function")
def _round(a, *args, **kwargs):
    val = np.round(a._nom, *args, **kwargs).squeeze()
    err = a._err + 0.5 * np.sign(a._err)
    return a.__class__(val, err)


@implements("take_along_axis", "function")
def _take_along_axis(a, *args, **kwargs):
    val = np.take_along_axis(a._nom, *args, **kwargs).squeeze()
    err = np.take_along_axis(a._err, *args, **kwargs).squeeze()
    return a.__class__(val, err)


@implements("concatenate", "function")
def _concatenate(concat, *args, **kwargs):
    for a in concat:
        if not _is_uncertainty(a):
            raise ValueError
    val = np.concatenate([a._nom for a in concat], **kwargs)
    err = np.concatenate([a._err for a in concat], **kwargs)
    return a.__class__(val, err)


@implements("ndim", "function")
def _ndim(a, *args, **kwargs):
    return np.ndim(a._nom)


def wrap_numpy(func_type, func, args, kwargs):
    """Return the result from a JAX+NumPy function/ufunc as wrapped by uncert."""

    if func_type == "function":
        handled = HANDLED_FUNCTIONS
        # Need to handle functions in submodules
        if isinstance(func, str):
            name = func
        else:
            name = ".".join(func.__module__.split(".")[1:] + [func.__name__])
    elif func_type == "ufunc":
        handled = HANDLED_UFUNCS
        # ufuncs do not have func.__module__
        if isinstance(func, str):
            name = func
        else:
            name = func.__name__
    else:
        raise ValueError("Invalid func_type {}".format(func_type))

    if name not in handled:
        return NotImplemented
    return handled[name](*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings

import hypothesis.strategies as st
import numpy as np
from hypothesis import given, settings
from hypothesis.extra import numpy as hnp

import auto_uncertainties

try:
    from pint import DimensionalityError
except ImportError:

    class DimensionalityError(Exception):
        pass


warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)


UNITS = [None]


def op_test(op, *args, **kwargs):
    with_unc = [a for a in args]
    without_unc = [a.value for a in args if hasattr(a, "_nom")]
    units = kwargs.pop("units", None)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        try:
            w_ = op(*without_unc, **kwargs)
        except DimensionalityError:
            with_unc = [a.m for a in args]
            without_unc = [a.value.m for a in args if hasattr(a, "_nom")]
            units = None
            w_ = op(*without_unc, **kwargs)
        except TypeError:
            return
        w = op(*with_unc, **kwargs)
        if units is not None and hasattr(w_, "units"):
            assert w.units == w_.units
            w = w.m
            w_ = w_.m
    if hasattr(w, "_nom"):
        np.testing.assert_almost_equal(w_, w.value, decimal=5)
    else:
        np.testing.assert_almost_equal(w_, w)
    return w


def given_float_3d(ops):
    def inner(func):
        return given(
            unom=hnp.arrays(
                dtype=st.sampled_from([np.float64]),
                shape=(2, 3),
                elements=st.floats(
                    min_value=-10.0,
                    max_value=10.0,
                    allow_nan=False,
                    allow_infinity=False,
                ),
            ),
            uerr=hnp.arrays(
                dtype=st.sampled_from([np.float64]),
                shape=(2, 3),
                elements=st.floats(
                    min_value=0.0,
                    max_value=10.0,
                    allow_nan=False,
                    allow_infinity=False,
                ),
            ),
            units=st.sampled_from(UNITS),
            op=st.sampled_from(ops),
        )(func)

    return inner


@given_float_3d(auto_uncertainties.wrap_numpy.bcast_same_shape_ufuncs)
@settings(deadline=3000)
def test_same_shape(unom, uerr, units, op):
    vnom = unom / 2
    verr = uerr / 2
    u = auto_uncertainties.Uncertainty(unom, uerr)
    v = auto_uncertainties.Uncertainty(vnom, verr)
    assert u.shape == v.shape
    if units is not None:
        u *= units
        v *= units
    oper = getattr(np, op)
    if op in auto_uncertainties.wrap_numpy.unary_bcast_same_shape_ufuncs:
        op_test(oper, u, units=units)
    else:
        op_test(oper, u, v, units=units)


@given_float_3d(auto_uncertainties.wrap_numpy.bcast_same_shape_bool_ufuncs)
@settings(deadline=3000)
def test_same_shape_bool(unom, uerr, units, op):
    vnom = unom / 2
    verr = uerr / 2
    u = auto_uncertainties.Uncertainty(unom, uerr)
    v = auto_uncertainties.Uncertainty(vnom, verr)
    if units is not None:
        u *= units
        v *= units
    oper = getattr(np, op)
    if op in auto_uncertainties.wrap_numpy.unary_bcast_same_shape_bool_ufuncs:
        w = oper(u)
    else:
        w = oper(u, v)
    assert w.dtype == bool


@given_float_3d(auto_uncertainties.wrap_numpy.bcast_nograd_ufuncs)
@settings(deadline=3000)
def test_nograd(unom, uerr, units, op):
    vnom = unom / 2
    verr = uerr / 2
    u = auto_uncertainties.Uncertainty(unom, uerr)
    v = auto_uncertainties.Uncertainty(vnom, verr)
    if units is not None:
        u *= units
        v *= units

    oper = getattr(np, op)
    try:
        w = op_test(oper, u, units=units)
    except TypeError:
        w = op_test(oper, u, v, units=units)
    assert not isinstance(w, auto_uncertainties.Uncertainty)


@given_float_3d(auto_uncertainties.wrap_numpy.bcast_apply_to_both_ufuncs)
@settings(deadline=3000)
def test_apply_to_both(unom, uerr, units, op):
    vnom = unom / 2
    verr = uerr / 2
    u = auto_uncertainties.Uncertainty(unom, uerr)
    v = auto_uncertainties.Uncertainty(vnom, verr)
    if units is not None:
        u *= units
        v *= units
    oper = getattr(np, op)
    op_test(oper, u, units=units)


@given_float_3d(auto_uncertainties.wrap_numpy.bcast_reduction_unary)
@settings(deadline=3000)
def test_unary_reduction(unom, uerr, units, op):
    vnom = unom / 2
    verr = uerr / 2
    u = auto_uncertainties.Uncertainty(unom, uerr)
    v = auto_uncertainties.Uncertainty(vnom, verr)
    if units is not None:
        u *= units
        v *= units
    oper = getattr(np, op)
    op_test(oper, u, units=units)



# -*- coding: utf-8 -*-
from __future__ import annotations

import copy
import re
from typing import Any, Sequence, Union

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from pandas.api.extensions import (
    ExtensionArray,
    ExtensionDtype,
    register_dataframe_accessor,
    register_extension_dtype,
)
from pandas.api.types import (
    is_integer,
    is_list_like,
    is_object_dtype,
    is_string_dtype,
)
from pandas.compat import set_function_name
from pandas.core.arrays.base import ExtensionOpsMixin
from pandas.core.indexers import check_array_indexer

from . import nominal_values, std_devs, Uncertainty


class UncertaintyType(ExtensionDtype):
    """
    An Uncertainty duck-typed class, suitable for holding an uncertainty (i.e. value and error pair) dtype. Closely follows the implementation in pint-pandas.
    """

    type = Uncertainty
    _metadata = ("value_dtype",)
    _match = re.compile(r"[U|u]ncertainty\[([a-zA-Z0-9]+)\]")
    _cache = {}
    value_dtype: Any

    @property
    def _is_numeric(self):
        # type: () -> bool
        return True

    def __new__(cls, value_dtype):
        """
        Parameters
        ----------
        units : Pint units or string
        """

        if isinstance(value_dtype, UncertaintyType):
            return value_dtype

        elif value_dtype is None:
            # empty constructor for pickle compat
            return object.__new__(cls)

        else:
            u = object.__new__(cls)
            u.value_dtype = np.dtype(value_dtype)

        return u

    @classmethod
    def construct_from_string(cls, string):
        """
        Strict construction from a string, raise a TypeError if not
        possible
        """
        if not isinstance(string, str):
            raise TypeError(
                f"'construct_from_string' expects a string, got {type(string)}"
            )
        try:
            dtype = np.dtype(str)
            return cls(value_dtype=dtype)
        except Exception:
            ...

        if cls._match.match(string):
            dtype = cls._match.match(string).group(1)
            return cls(value_dtype=dtype)

        if isinstance(string, str) and (
            string.startswith("uncertainty[") or string.startswith("Pint[")
        ):
            # do not parse string like U as pint[U]
            # avoid tuple to be regarded as unit
            try:
                return cls(units=string)
            except ValueError:
                pass
        raise TypeError(
            f"Cannot construct a 'UncertaintyType' from '{string}'"
        )

    @property
    def name(self):
        return f"Uncertainty[{self.value_dtype}]"

    @property
    def na_value(self):
        return Uncertainty(np.nan, np.nan)

    def __hash__(self):
        # make myself hashable
        return hash(str(self))

    def __eq__(self, other):
        try:
            other = UncertaintyType(other)
        except ValueError:
            return False
        return self.value_dtype == other.value_dtype

    @classmethod
    def is_dtype(cls, dtype):
        """
        Return a boolean if we if the passed type is an actual dtype that we
        can match (via string or type)
        """
        if isinstance(dtype, str):
            return cls._match.match(dtype)
        return super(UncertaintyType, cls).is_dtype(dtype)

    @classmethod
    def construct_array_type(cls):
        return UncertaintyArray

    def __repr__(self):
        """
        Return a string representation for this object.

        Invoked by unicode(df) in py2 only. Yields a Unicode String in both
        py2/py3.
        """

        return self.name


dtypemap = {
    int: pd.Int64Dtype(),
    np.int64: pd.Int64Dtype(),
    np.int32: pd.Int32Dtype(),
    np.int16: pd.Int16Dtype(),
    np.int8: pd.Int8Dtype(),
    # np.float128: pd.Float128Dtype(),
    float: pd.Float64Dtype(),
    np.float64: pd.Float64Dtype(),
    np.float32: pd.Float32Dtype(),
    np.complex128: pd.core.dtypes.dtypes.PandasDtype("complex128"),
    np.complex64: pd.core.dtypes.dtypes.PandasDtype("complex64"),
    # np.float16: pd.Float16Dtype(),
}
dtypeunmap = {v: k for k, v in dtypemap.items()}


class UncertaintyArray(ExtensionArray, ExtensionOpsMixin):
    """Implements a class to describe an array of physical quantities:
    the product of an array of numerical values and a unit of measurement.

    Parameters
    ----------
    values : scalar or array-like
        Array of mean/nominal values
    errors: scalar or array-like
        Array of standard deviations/uncertainties
    dtype : UncertaintyType or str
        Datatype of the underlying values
    copy: bool
        Whether to copy the values.
    Returns
    -------

    """

    _data = np.array([])
    context_name = None

    def __init__(
        self,
        values,
        errors=None,
        dtype=None,
        copy=False,
    ):
        if dtype is None:
            if isinstance(values, np.ndarray):
                dtype = UncertaintyType(value_dtype=values.dtype)
        if isinstance(dtype, str):
            dtype = UncertaintyType.construct_from_string(dtype)
        if not isinstance(dtype, UncertaintyType):
            raise NotImplementedError

        self._dtype = dtype

        if isinstance(values, (UncertaintyArray)):
            values = values._nom
            errors = values._err
        elif isinstance(values, (Uncertainty)):
            values = values.value
            errors = values.error
        else:
            assert (
                errors is not None
            ), "errors must be specified for non UncertaintyArray values"

        if isinstance(values, np.ndarray):
            dtype = values.dtype
            if dtype in dtypemap:
                dtype = dtypemap[dtype]
            values = pd.array(values, copy=copy, dtype=dtype)
            errors = pd.array(errors, copy=copy, dtype=dtype)
            copy = False
        elif not isinstance(values, pd.core.arrays.numeric.NumericArray):
            values = pd.array(values, copy=copy)
            errors = pd.array(errors, copy=copy)
            copy = False
        if copy:
            values = values.copy()
            errors = errors.copy()

        self._nom = values
        self._err = errors

    def __getstate__(self):
        # we need to discard the cached _Q, which is not pickleable
        ret = dict(self.__dict__)
        return ret

    def __setstate__(self, dct):
        self.__dict__.update(dct)

    @property
    def dtype(self):
        # type: () -> ExtensionDtype
        """An instance of 'ExtensionDtype'."""
        return self._dtype

    def __len__(self):
        # type: () -> int
        """Length of this array

        Returns
        -------
        length : int
        """
        return len(self._nom)

    def __getitem__(self, item):
        # type (Any) -> Any
        """Select a subset of self.
        Parameters
        ----------
        item : int, slice, or ndarray
            * int: The position in 'self' to get.
            * slice: A slice object, where 'start', 'stop', and 'step' are
              integers or None
            * ndarray: A 1-d boolean NumPy ndarray the same length as 'self'
        Returns
        -------
        item : scalar Uncertainty or UncertaintyArray
        """

        if is_integer(item):
            return Uncertainty(self._nom[item], self._err[item])

        item = check_array_indexer(self, item)

        return self.__class__(
            self._nom[item],
            self._err[item],
        )

    def __setitem__(
        self,
        key,
        value: Union[Uncertainty, UncertaintyArray, Sequence[Uncertainty]],
    ):
        # need to not use `not value` on numpy arrays
        if isinstance(value, (list, tuple)) and (not value):
            # doing nothing here seems to be ok
            return

        if isinstance(value, self._dtype.type):
            val = value.value
            err = value.error

        elif (
            is_list_like(value)
            and len(value) > 0
            and isinstance(value[0], self._dtype.type)
        ):
            val = [item.value for item in value]
            err = [item.error for item in value]

        key = check_array_indexer(self, key)
        try:
            self._nom[key] = val
            self._err[key] = err
        except IndexError as e:
            msg = "Mask is wrong length. {}".format(e)
            raise IndexError(msg)

    def _formatter(self, boxed=False):
        """Formatting function for scalar values.
        This is used in the default '__repr__'. The returned formatting
        function receives scalar Uncertainties.

        # type: (bool) -> Callable[[Any], Optional[str]]

        Parameters
        ----------
        boxed: bool, default False
            An indicated for whether or not your array is being printed
            within a Series, DataFrame, or Index (True), or just by
            itself (False). This may be useful if you want scalar values
            to appear differently within a Series versus on its own (e.g.
            quoted or not).

        Returns
        -------
        Callable[[Any], str]
            A callable that gets instances of the scalar type and
            returns a string. By default, :func:`repr` is used
            when ``boxed=False`` and :func:`str` is used when
            ``boxed=True``.
        """

        def formatting_function(uncertainty: Uncertainty):
            return str(uncertainty)

        return formatting_function

    def isna(self):
        # type: () -> np.ndarray
        """Return a Boolean NumPy array indicating if each value is missing.

        Returns
        -------
        missing : np.array
        """
        return ~np.isfinite(self._nom)

    def astype(self, dtype, copy=True):
        """Cast to a NumPy array with 'dtype'.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, default True
            Whether to copy the data, even if not necessary. If False,
            a copy is made only if the old dtype does not match the
            new dtype.

        Returns
        -------
        array : ndarray
            NumPy ndarray with 'dtype' for its dtype.
        """
        if isinstance(dtype, str) and self._dtype._match(dtype):
            dtype = self._dtype.construct_from_string(dtype)

        if isinstance(dtype, self._dtype):
            if dtype == self._dtype and not copy:
                return self
            else:
                return UncertaintyArray(
                    self.uncertainty.to(dtype.value_dtype), dtype
                )

        # do *not* delegate to __array__ -> is required to return a numpy array,
        # but somebody may be requesting another pandas array
        # examples are e.g. PyArrow arrays as requested by "string[pyarrow]"
        if is_object_dtype(dtype):
            return self._to_array_of_quantity(copy=copy)
        if is_string_dtype(dtype):
            return pd.array([str(x) for x in self.uncertainty], dtype=dtype)
        return pd.array(self.uncertainty, dtype, copy)

    @property
    def uncertainty(self):
        return Uncertainty(self._nom, self._err)

    def take(
        self,
        indices: Sequence[int],
        allow_fill: bool = False,
        fill_value: Uncertainty = None,
    ):
        """Take elements from an array.

        # type: (Sequence[int], bool, Optional[Any]) -> UncertaintyArray

        Parameters
        ----------
        indices : sequence of integers
            Indices to be taken.
        allow_fill : bool, default False
            How to handle negative values in `indices`.
            * False: negative values in `indices` indicate positional indices
              from the right (the default). This is similar to
              :func:`numpy.take`.
            * True: negative values in `indices` indicate
              missing values. These values are set to `fill_value`. Any other
              other negative values raise a ``ValueError``.
        fill_value : any, optional
            Fill value to use for NA-indices when `allow_fill` is True.
            This may be ``None``, in which case the default NA value for
            the type, ``self.dtype.na_value``, is used.

        Returns
        -------
        UncertaintyArray

        Raises
        ------
        IndexError
            When the indices are out of bounds for the array.
        ValueError
            When `indices` contains negative values other than ``-1``
            and `allow_fill` is True.
        Notes
        -----
        UncertaintyArray.take is called by ``Series.__getitem__``, ``.loc``,
        ``iloc``, when `indices` is a sequence of values. Additionally,
        it's called by :meth:`Series.reindex`, or any other method
        that causes realignemnt, with a `fill_value`.
        See Also
        --------
        numpy.take
        pandas.api.extensions.take
        Examples
        --------
        """
        from pandas.core.algorithms import take

        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value
        else:
            fill_value = Uncertainty(0, 0)

        value = take(
            self._nom,
            indices,
            fill_value=float(fill_value.value),
            allow_fill=allow_fill,
        )
        error = take(
            self._err,
            indices,
            fill_value=float(fill_value.error),
            allow_fill=allow_fill,
        )

        return UncertaintyArray(value, error, dtype=self.dtype)

    def copy(self, deep: bool = False):
        data = (self._nom, self._err)
        if deep:
            data = copy.deepcopy(data)
        else:
            data = copy.copy(data)

        return type(self)(*data, dtype=self.dtype)

    @classmethod
    def _concat_same_type(cls, to_concat: Sequence[UncertaintyArray]):
        v = []
        e = []
        dtype = None
        for a in to_concat:
            if dtype is None:
                dtype = a.dtype
            else:
                a = a.astype(dtype)
            v.append(a._nom)
            e.append(a._err)

        return cls(np.concatenate(v), np.concatenate(e), dtype)

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        """
        Initialises a UncertaintyArray from a list like of Uncertainty scalars or a list like of value/error and dtype
        -----
        Usage
        UncertaintyArray._from_sequence([Uncertainty(1,0),Uncertainty(1,1)])
        """

        list_of_scalars = []
        for s in scalars:
            if isinstance(s, Uncertainty):
                list_of_scalars.append(s)
            elif isinstance(s, tuple):
                list_of_scalars.append(Uncertainty(*s))
            else:
                list_of_scalars.append(Uncertainty(s, 0))

        values = np.asarray([x.value for x in list_of_scalars])
        errors = np.asarray([x.error for x in list_of_scalars])

        return cls(values, errors, dtype=dtype, copy=copy)

    @classmethod
    def _from_sequence_of_strings(cls, scalars, dtype=None, copy=False):
        list_of_uncs = [Uncertainty.from_string(x) for x in scalars]
        return cls._from_sequence(list_of_uncs, dtype=dtype, copy=copy)

    def value_counts(self, dropna=True):
        """
        Returns a Series containing counts of each category.

        Every category will have an entry, even those with a count of 0.

        Parameters
        ----------
        dropna : boolean, default True
            Don't include counts of NaN.

        Returns
        -------
        counts : Series

        See Also
        --------
        Series.value_counts
        """

        from pandas import Series

        # compute counts on the data with no nans
        data = Uncertainty(self._nom, self._err)

        nafilt = np.isnan(data)
        data = data[~nafilt]

        data_list = [Uncertainty(x.value, x.error) for x in data]
        index = list(set(data))
        array = [data_list.count(item) for item in index]

        if not dropna:
            index.append(np.nan)
            array.append(nafilt.sum())

        return Series(array, index=index)

    def unique(self):
        """Compute the UncertaintyArray of unique values.

        Returns
        -------
        uniques : UncertaintyArray
        """
        from pandas import unique

        data = Uncertainty(self._nom, self._err)

        return self._from_sequence(unique(data), dtype=self.dtype)

    def __contains__(self, item) -> bool:
        if not isinstance(item, Uncertainty):
            return False
        elif pd.isna(item):
            return self.isna().any()
        else:
            return super().__contains__(item)

    @property
    def nbytes(self):
        return self._nom.nbytes * 2

    # The _can_hold_na attribute is set to True so that pandas internals
    # will use the ExtensionDtype.na_value as the NA value in operations
    # such as take(), reindex(), shift(), etc.  In addition, those results
    # will then be of the ExtensionArray subclass rather than an array
    # of objects
    _can_hold_na = True

    @property
    def _ndarray_values(self):
        # type: () -> np.ndarray
        """Internal pandas method for lossy conversion to a NumPy ndarray.
        This method is not part of the pandas interface.
        The expectation is that this is cheap to compute, and is primarily
        used for interacting with our indexers.
        """
        return np.array(self)

    @classmethod
    def _create_method(cls, op, coerce_to_dtype=True):
        """
        A class method that returns a method that will correspond to an
        operator for an ExtensionArray subclass, by dispatching to the
        relevant operator defined on the individual elements of the
        ExtensionArray.
        Parameters
        ----------
        op : function
            An operator that takes arguments op(a, b)
        coerce_to_dtype :  bool
            boolean indicating whether to attempt to convert
            the result to the underlying ExtensionArray dtype
            (default True)
        Returns
        -------
        A method that can be bound to a method of a class
        Example
        -------
        Given an ExtensionArray subclass called MyExtensionArray, use
        >>> __add__ = cls._create_method(operator.add)
        in the class definition of MyExtensionArray to create the operator
        for addition, that will be based on the operator implementation
        of the underlying elements of the ExtensionArray
        """

        def _binop(self: UncertaintyArray, other):
            def validate_length(obj1, obj2):
                # validates length
                try:
                    if len(obj1) != len(obj2):
                        raise ValueError("Lengths must match")
                except TypeError:
                    pass

            def convert_values(param):
                # convert to a quantity or listlike
                if isinstance(param, cls):
                    return param.uncertainty
                elif isinstance(param, Uncertainty):
                    return param
                elif (
                    is_list_like(param)
                    and len(param) > 0
                    and isinstance(param[0], Uncertainty)
                ):
                    return param[0]
                else:
                    return param

            if isinstance(other, (Series, DataFrame)):
                return NotImplemented
            lvalues = self.uncertainty
            validate_length(lvalues, other)
            rvalues = convert_values(other)

            # If the operator is not defined for the underlying objects,
            # a TypeError should be raised
            res = op(lvalues, rvalues)

            if coerce_to_dtype:
                try:
                    val = nominal_values(res)
                    err = std_devs(res)
                    res = cls(val, err, dtype=self.dtype)
                except TypeError:
                    pass

            return res

        op_name = f"__{op}__"
        return set_function_name(_binop, op_name, cls)

    @classmethod
    def _create_arithmetic_method(cls, op):
        return cls._create_method(op)

    @classmethod
    def _create_comparison_method(cls, op):
        return cls._create_method(op, coerce_to_dtype=False)

    def __array__(self, dtype=None, copy=False):
        if dtype is None or is_object_dtype(dtype):
            return self.uncertainty
        if is_string_dtype(dtype):
            return np.array([str(x) for x in self.uncertainty], dtype=str)
        return Uncertainty(self._nom.astype(dtype), self._err.astype(dtype))

    def searchsorted(self, value, side="left", sorter=None):
        """
        Find indices where elements should be inserted to maintain order.

        .. versionadded:: 0.24.0

        Find the indices into a sorted array `self` (a) such that, if the
        corresponding elements in `v` were inserted before the indices, the
        order of `self` would be preserved.

        Assuming that `a` is sorted:

        ======  ============================
        `side`  returned index `i` satisfies
        ======  ============================
        left    ``self[i-1] < v <= self[i]``
        right   ``self[i-1] <= v < self[i]``
        ======  ============================

        Parameters
        ----------
        value : array_like
            Values to insert into `self`.
        side : {'left', 'right'}, optional
            If 'left', the index of the first suitable location found is given.
            If 'right', return the last such index.  If there is no suitable
            index, return either 0 or N (where N is the length of `self`).
        sorter : 1-D array_like, optional
            Optional array of integer indices that sort array a into ascending
            order. They are typically the result of argsort.

        Returns
        -------
        indices : array of ints
            Array of insertion points with the same shape as `value`.

        See Also
        --------
        numpy.searchsorted : Similar method from NumPy.
        """
        # Note: the base tests provided by pandas only test the basics.
        # We do not test
        # 1. Values outside the range of the `data_for_sorting` fixture
        # 2. Values between the values in the `data_for_sorting` fixture
        # 3. Missing values.

        arr = self._nom
        if isinstance(value, Uncertainty):
            val = value.value
        elif (
            is_list_like(value)
            and len(value) > 0
            and isinstance(value[0], Uncertainty)
        ):
            val = [item.value for item in value]
        return arr.searchsorted(val, side=side, sorter=sorter)

    def _reduce(self, name, **kwds):
        """
        Return a scalar result of performing the reduction operation.

        Parameters
        ----------
        name : str
            Name of the function, supported values are:
            { any, all, min, max, sum, mean, median, prod,
            std, var, sem, kurt, skew }.
        skipna : bool, default True
            If True, skip NaN values.
        **kwargs
            Additional keyword arguments passed to the reduction function.
            Currently, `ddof` is the only supported kwarg.

        Returns
        -------
        scalar

        Raises
        ------
        TypeError : subclass does not define reductions
        """

        functions = {
            "min": np.nanmin,
            "max": np.nanmax,
            "sum": np.nansum,
            "mean": np.nanmean,
            "median": np.nanmedian,
            "std": np.nanstd,
            "var": np.nanvar,
        }
        if name not in functions:
            raise TypeError(f"cannot perform {name} with type {self.dtype}")

        result = functions[name](self.uncertainty, **kwds)
        return result


UncertaintyArray._add_arithmetic_ops()
UncertaintyArray._add_comparison_ops()
register_extension_dtype(UncertaintyType)


@register_dataframe_accessor("uncertainty")
class UncertaintyDataFrameAccessor(object):
    def __init__(self, pandas_obj: pd.DataFrame):
        self._obj = pandas_obj


class UncertaintySeriesAccessor(object):
    def __init__(self, pandas_obj: pd.Series):
        self._validate(pandas_obj)
        self.pandas_obj = pandas_obj
        self.uncertainty = pandas_obj.values
        self._index = pandas_obj.index
        self._name = pandas_obj.name

    @staticmethod
    def _validate(obj):
        if not is_uncert_type(obj):
            raise AttributeError(
                "Cannot use 'uncertainty' accessor on objects of "
                "dtype '{}'.".format(obj.dtype)
            )


class Delegated:
    # Descriptor for delegating attribute access to from
    # a Series to an underlying array
    to_series = True

    def __init__(self, name):
        self.name = name


class DelegatedProperty(Delegated):
    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, "_index")
        name = object.__getattribute__(obj, "_name")
        result = getattr(
            object.__getattribute__(obj, "uncertainty"), self.name
        )
        if self.to_series:
            if isinstance(result, Uncertainty):
                result = UncertaintyArray(result)
            return Series(result, index, name=name)
        else:
            return result


class DelegatedScalarProperty(DelegatedProperty):
    to_series = False


class DelegatedMethod(Delegated):
    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, "_index")
        name = object.__getattribute__(obj, "_name")
        method = getattr(
            object.__getattribute__(obj, "uncertainty"), self.name
        )

        def delegated_method(*args, **kwargs):
            result = method(*args, **kwargs)
            if self.to_series:
                if isinstance(result, Uncertainty):
                    result = UncertaintyArray(result)
                result = Series(result, index, name=name)
            return result

        return delegated_method


class DelegatedScalarMethod(DelegatedMethod):
    to_series = False


for attr in [
    "debug_used",
    "default_format",
    "dimensionality",
    "dimensionless",
    "force_ndarray",
    "shape",
    "u",
    "unitless",
    "units",
]:
    setattr(UncertaintySeriesAccessor, attr, DelegatedScalarProperty(attr))
for attr in ["imag", "m", "magnitude", "real"]:
    setattr(UncertaintySeriesAccessor, attr, DelegatedProperty(attr))

for attr in [
    "check",
    "compatible_units",
    "format_babel",
    "ito",
    "ito_base_units",
    "ito_reduced_units",
    "ito_root_units",
    "plus_minus",
    "put",
    "to_tuple",
    "tolist",
]:
    setattr(UncertaintySeriesAccessor, attr, DelegatedScalarMethod(attr))
for attr in [
    "clip",
    "from_tuple",
    "m_as",
    "searchsorted",
    "to",
    "to_base_units",
    "to_compact",
    "to_reduced_units",
    "to_root_units",
    "to_timedelta",
]:
    setattr(UncertaintySeriesAccessor, attr, DelegatedMethod(attr))


def is_uncert_type(obj):
    t = getattr(obj, "dtype", obj)
    try:
        return isinstance(t, UncertaintyType) or issubclass(t, UncertaintyType)
    except Exception:
        return False


# try:
#     # for pint < 0.21 we need to explicitly register
#     compat.upcast_types.append(UncertaintyArray)
# except AttributeError:
#     # for pint = 0.21 we need to add the full names of UncertaintyArray and DataFrame,
#     # which is to be added in pint > 0.21
#     compat.upcast_type_map.setdefault("pint_pandas.pint_array.UncertaintyArray", UncertaintyArray)
#     compat.upcast_type_map.setdefault("pandas.core.frame.DataFrame", DataFrame)



# -*- coding: utf-8 -*-
from __future__ import annotations

import operator

import numpy as np
import pandas as pd
import pytest
from pandas.tests.extension.conftest import as_array  # noqa: F401
from pandas.tests.extension.conftest import as_frame  # noqa: F401
from pandas.tests.extension.conftest import as_series  # noqa: F401
from pandas.tests.extension.conftest import fillna_method  # noqa: F401
from pandas.tests.extension.conftest import groupby_apply_op  # noqa: F401
from pandas.tests.extension.conftest import use_numpy  # noqa: F401

from auto_uncertainties.pandas_compat import UncertaintyArray


@pytest.fixture
def data():
    return UncertaintyArray(
        np.random.random(size=100), np.abs(np.random.random(size=100))
    )


class TestUserInterface(object):
    def test_get_underlying_data(self, data: UncertaintyArray):
        ser = pd.Series(data)
        # this first test creates an array of bool (which is desired, eg for indexing)
        assert all(ser.values == data)
        assert ser.values[23] == data[23]

    def test_arithmetic(self, data: UncertaintyArray):
        ser = pd.Series(data)
        ser2 = ser + ser
        assert all(ser2.values == 2 * data)

    def test_initialisation(self, data: UncertaintyArray):
        # fails with plain array
        # works with UncertaintyArray
        df = pd.DataFrame(
            {
                "x1": pd.Series(data.uncertainty, dtype="Uncertainty[float]"),
                "x2": pd.Series(data.uncertainty.value, dtype="float"),
                "x3": pd.Series(
                    data.uncertainty.value,
                    data.uncertainty.error,
                    dtype="uncertainty[float32]",
                ),
            }
        )

        for col in df.columns:
            assert all(df[col] == df.length)

    def test_df_operations(self):
        # simply a copy of what's in the notebook
        df = pd.DataFrame(
            {
                "torque": pd.Series(
                    [1.0, 2.0, 2.0, 3.0], dtype="Uncertainty[float]"
                ),
                "angular_velocity": UncertaintyArray(
                    [1.0, 2.0, 2.0, 3.0],
                    [1.0, 2.0, 2.0, 3.0],
                    dtype="Uncertainty[float]",
                ),
            }
        )

        df["power"] = df["torque"] * df["angular_velocity"]

        df.power.values
        df.power.values.value
        df.power.values.error

        df.angular_velocity.values


arithmetic_ops = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.truediv,
    operator.floordiv,
    operator.pow,
]

comparative_ops = [
    operator.eq,
    operator.le,
    operator.lt,
    operator.ge,
    operator.gt,
]

unit_ops = [
    operator.mul,
    operator.truediv,
]



# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from functools import wraps

import numpy as np

from . import NumpyDowncastWarning


def ignore_runtime_warnings(f):
    """
    A decorator to ignore runtime warnings
    Parameters
    ----------
    f: function
        The wrapped function

    Returns
    -------
    wrapped_function: function
        The wrapped function
    """

    @wraps(f)
    def runtime_warn_inner(*args, **kwargs):
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore", category=RuntimeWarning)
            response = f(*args, **kwargs)
        return response

    return runtime_warn_inner


def ignore_numpy_downcast_warnings(f):
    """
    A decorator to ignore NumpyDowncastWarning
    Parameters
    ----------
    f: function
        The wrapped function

    Returns
    -------
    wrapped_function: function
        The wrapped function
    """

    @wraps(f)
    def user_warn_inner(*args, **kwargs):
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore", category=NumpyDowncastWarning)
            response = f(*args, **kwargs)
        return response

    return user_warn_inner


def is_iterable(y):
    try:
        iter(y)
    except TypeError:
        return False
    return True


def has_length(y):
    try:
        len(y)
    except TypeError:
        return False
    return True


def is_np_duck_array(cls):
    """Check if object is a numpy array-like, but not a Uncertainty

    Parameters
    ----------
    cls : class

    Returns
    -------
    bool
    """
    try:
        import numpy as np
    except ImportError:
        return False

    return issubclass(cls, np.ndarray) or (
        not hasattr(cls, "_nom")
        and not hasattr(cls, "_err")
        and hasattr(cls, "__array_function__")
        and hasattr(cls, "ndim")
        and hasattr(cls, "dtype")
    )


class Display(object):
    default_format: str = ""

    def _repr_html_(self):
        val_ = self._nom
        err_ = self._err
        if is_np_duck_array(type(self._nom)):
            header = "<table><tbody>"
            footer = "</tbody></table>"
            val = f"<tr><th>Magnitude</th><td style='text-align:left;'><pre>{val_}</pre></td></tr>"
            err = f"<tr><th>Error</th><td style='text-align:left;'><pre>{err_}</pre></td></tr>"
            return header + val + err + footer
        else:
            val = f"{val_}"
            err = f"{err_}"
            return f"{val} {chr(0x00B1)} {err}"

    def _repr_latex_(self):
        val_ = self._nom
        err_ = self._err
        if is_np_duck_array(type(self._nom)):
            s = (
                ", ".join(
                    [
                        f"{v} \\pm {e}"
                        for v, e in zip(val_.ravel(), err_.ravel())
                    ]
                )
                + "~"
            )
            header = "$"
            footer = "$"
            return header + s + footer
        else:
            val = f"{val_}"
            err = f"{err_}"
            return f"${val} \\pm {err}$"

    def __str__(self) -> str:
        val_ = self._nom
        err_ = self._err

        if self._nom is not None:
            if self._err is not None:
                if is_np_duck_array(type(self._nom)):
                    return (
                        "["
                        + ", ".join(
                            [
                                f"{v} +/- {e}"
                                for v, e in zip(val_.ravel(), err_.ravel())
                            ]
                        )
                        + "]"
                    )
                else:
                    return f"{val_} +/- {err_}"
            else:
                return f"{val_}"

    def __format__(self, fmt):
        val_ = self._nom
        err_ = self._err
        str_rep = f"{val_:{fmt}} +/- {err_:{fmt}}"
        return str_rep

    def __repr__(self) -> str:
        return str(self)


def strip_device_array(value):
    return np.array(value)


def ndarray_to_scalar(value):
    return np.ndarray.item(strip_device_array(value))



# -*- coding: utf-8 -*-
from __future__ import annotations

import operator
import warnings

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra import numpy as hnp

from auto_uncertainties import NegativeStdDevError, Uncertainty

try:
    from pint import DimensionalityError
except ImportError:

    class DimensionalityError(Exception):
        pass


BINARY_OPS = [
    operator.lt,
    operator.le,
    operator.eq,
    operator.ne,
    operator.gt,
    operator.ge,
]
UNARY_OPS = [operator.not_, operator.abs]
warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
UNITS = [None]


def check_units_and_mag(unc, units, mag, err):
    if units is not None and not hasattr(unc, "units"):
        raise ValueError
    if units is None:
        assert unc.value == mag
        assert unc.error == err
    else:
        assert unc.units.is_compatible_with(units)
        assert unc.value.units.is_compatible_with(units)
        assert unc.value.to(units).m == mag
        assert unc.error.units.is_compatible_with(units)
        assert unc.error.to(units).m == err


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v=st.floats(allow_subnormal=False),
    e=st.floats(allow_subnormal=False),
    units=st.sampled_from(UNITS),
)
def test_scalar_creation(v, e, units):
    if e < 0:
        if np.isfinite(e):
            with pytest.raises(NegativeStdDevError):
                u = Uncertainty(v, e)
            if units is not None:
                with pytest.raises(NegativeStdDevError):
                    u = Uncertainty.from_quantities(v * units, e * units)
    else:
        u = Uncertainty(v, e)
        if np.isfinite(v) and np.isfinite(e):
            assert u.value == v
            assert u.error == e
            if v > 0:
                assert u.relative == e / v
            elif v == 0:
                assert not np.isfinite(u.relative)

            if units is not None:
                with pytest.raises(NotImplementedError):
                    u = Uncertainty(v * units, e * units)

                u = Uncertainty.from_quantities(v * units, e * units)
                check_units_and_mag(u, units, v, e)

                u = Uncertainty(v, e) * units
                check_units_and_mag(u, units, v, e)

                with pytest.raises(DimensionalityError):
                    u = Uncertainty.from_quantities(v * units, e)
                with pytest.raises(DimensionalityError):
                    u = Uncertainty.from_quantities(v, e * units)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v1=st.floats(allow_subnormal=False),
    e1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    op=st.sampled_from(UNARY_OPS),
)
def test_scalar_unary(v1, e1, op):
    u1 = Uncertainty(v1, e1)

    u = op(u1)
    if np.isfinite(v1):
        if isinstance(u, Uncertainty):
            assert u.value == op(u1.value)
            if np.isfinite(e1):
                assert np.isfinite(u.error)
        else:
            assert u == op(u1.value)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v1=st.floats(allow_subnormal=False),
    v2=st.floats(allow_subnormal=False),
    e1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    e2=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    op=st.sampled_from(BINARY_OPS),
)
def test_scalar_binary(v1, e1, v2, e2, op):
    u1 = Uncertainty(v1, e1)
    u2 = Uncertainty(v2, e2)

    u = op(u1, u2)
    if isinstance(u, Uncertainty):
        assert u.value == op(u1.value, u2.value)
        if np.isfinite(e1) and np.isfinite(e2):
            assert np.isfinite(u.error)
    else:
        assert u == op(u1.value, u2.value)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v1=st.floats(allow_subnormal=False),
    v2=st.floats(allow_subnormal=False),
    e1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    e2=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    op=st.sampled_from([operator.add, operator.sub]),
)
def test_scalar_add_sub(v1, e1, v2, e2, op):
    u1 = Uncertainty(v1, e1)
    u2 = Uncertainty(v2, e2)

    u = op(u1, u2)
    if np.isfinite(v1) and np.isfinite(v2):
        assert u.value == op(u1.value, u2.value)
    if np.isfinite(e1) and np.isfinite(e2):
        assert u.error == np.sqrt(u1.error**2 + u2.error**2)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v1=st.floats(min_value=1, max_value=1e3, allow_subnormal=False),
    v2=st.floats(min_value=1, max_value=1e3, allow_subnormal=False),
    e1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    e2=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    op=st.sampled_from([operator.mul, operator.truediv]),
)
def test_scalar_mul_div(v1, e1, v2, e2, op):
    u1 = Uncertainty(v1, e1)
    u2 = Uncertainty(v2, e2)

    u = op(u1, u2)
    if np.isfinite(v1) and np.isfinite(v2):
        assert u.value == op(u1.value, u2.value)
    if np.isfinite(e1) and np.isfinite(e2):
        np.testing.assert_almost_equal(
            u.error, u.value * np.sqrt(u1.rel**2 + u2.rel**2)
        )


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@given(
    v1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    v2=hnp.arrays(
        dtype=st.sampled_from([np.float64]),
        shape=(11,),
        elements=st.floats(
            min_value=-10.0,
            max_value=10.0,
            allow_nan=False,
            allow_infinity=False,
        ),
    ),
    e1=st.floats(min_value=0, max_value=1e3, allow_subnormal=False),
    e2=hnp.arrays(
        dtype=st.sampled_from([np.float64]),
        shape=(11,),
        elements=st.floats(
            min_value=0, max_value=10.0, allow_nan=False, allow_infinity=False
        ),
    ),
    op=st.sampled_from(
        [
            operator.add,
            operator.sub,
            operator.mul,
            operator.truediv,
            operator.mod,
            operator.pow,
        ]
    ),
)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_mixed_arithmetic(v1, e1, v2, e2, op):
    u1 = Uncertainty(v1, e1)
    u2 = Uncertainty(v2, e2)

    u = op(u1, u2)
    np.testing.assert_almost_equal(u.value, op(u1.value, u2.value))

    u = op(u2, u1)
    np.testing.assert_almost_equal(u.value, op(u2.value, u1.value))



# -*- coding: utf-8 -*-
# Based heavily on the implementation of pint's Quantity object
from __future__ import annotations

import copy
import locale
import operator
import warnings

import joblib
import numpy as np

from . import NegativeStdDevError, NumpyDowncastWarning
from .util import (
    Display,
    ignore_numpy_downcast_warnings,
    ignore_runtime_warnings,
    is_np_duck_array,
    strip_device_array,
)
from .wrap_numpy import HANDLED_FUNCTIONS, HANDLED_UFUNCS, wrap_numpy


def _check_units(value, err):
    mag_has_units = hasattr(value, "units")
    mag_units = getattr(value, "units", None)
    err_has_units = hasattr(err, "units")
    err_units = getattr(err, "units", None)

    if mag_has_units and mag_units is not None:
        Q = mag_units._REGISTRY.Quantity
        ret_val = Q(value).to(mag_units).m
        if err is not None:
            ret_err = Q(err).to(mag_units).m
        else:
            ret_err = None
        ret_units = mag_units
    # This branch will never actually work, but its here
    # to raise a Dimensionality error without needing to import pint
    elif err_has_units:
        Q = err_units._REGISTRY.Quantity
        ret_val = Q(value).to(err_units).m
        ret_err = Q(err).to(err_units).m
        ret_units = err_units
    else:
        ret_units = None
        ret_val = value
        ret_err = err

    return ret_val, ret_err, ret_units


class Uncertainty(Display):
    __apply_to_both_ndarray__ = [
        "flatten",
        "real",
        "imag",
        "astype",
        "T",
        "reshape",
    ]
    __ndarray_attributes__ = ["dtype", "ndim", "size"]

    __array_priority__ = 18

    @ignore_numpy_downcast_warnings
    def __init__(self, value, err=None):
        if hasattr(value, "units") or hasattr(err, "units"):
            raise NotImplementedError(
                "Uncertainty cannot have units! Call Uncertainty.from_quantities instead."
            )

        value_ = strip_device_array(value)
        if err is not None:
            err_ = strip_device_array(err)
        else:
            err_ = None

        # If Uncertatity
        if isinstance(value_, self.__class__):
            magnitude_nom = value_.value
            magnitude_err = value_.error
        # If sequence
        elif isinstance(value_, list):
            inst = self.__class__.from_list(value_)
            magnitude_nom = inst.value
            magnitude_err = inst.error
        # If arrays
        elif np.ndim(value_) > 0:
            magnitude_nom = np.asarray(value_)
            if err_ is None:
                magnitude_err = np.zeros_like(value_)
            else:
                if np.ndim(err_) == 0:
                    magnitude_err = np.ones_like(value_) * err_
                else:
                    magnitude_err = np.asarray(err_)
                    assert magnitude_err.shape == magnitude_nom.shape
        # If scalar
        else:
            magnitude_nom = value_
            if err_ is None:
                magnitude_err = 0.0
            else:
                magnitude_err = err_

        # Replace NaNs in errors with zeros
        if is_np_duck_array(type(magnitude_err)):
            magnitude_err[~np.isfinite(magnitude_err)] = 0
        else:
            if not np.isfinite(magnitude_err):
                magnitude_err = 0

        # Basic sanity checks
        if is_np_duck_array(type(magnitude_nom)):
            match_items = self.__ndarray_attributes__ + ["shape"]
            try:
                match_items.remove("dtype")
            except ValueError:
                pass
            for item in match_items:
                if not getattr(magnitude_nom, item) == getattr(
                    magnitude_err, item
                ):
                    raise ValueError(
                        f"Attribute {item} does not match for value and error! ({getattr(magnitude_nom,item)} and {getattr(magnitude_err,item)})"
                    )
        err_mag = np.atleast_1d(magnitude_err)
        if np.any(err_mag[np.isfinite(err_mag)] < 0):
            valid = err_mag[np.isfinite(err_mag)]

            raise NegativeStdDevError(
                f"Found negatives values for the standard deviation... {valid[valid < 0]}"
            )

        self._nom = magnitude_nom
        self._err = magnitude_err

    def __bytes__(self) -> bytes:
        return str(self).encode(locale.getpreferredencoding())

    def __iter__(self):
        for v, e in zip(self._nom, self._err):
            yield self.__class__(v, e)

    def __copy__(self) -> Uncertainty:
        ret = self.__class__(copy.copy(self._nom), copy.copy(self._err))

        return ret

    def __deepcopy__(self, memo) -> Uncertainty:
        ret = self.__class__(
            copy.deepcopy(self._nom, memo), copy.deepcopy(self._err, memo)
        )
        return ret

    def __hash__(self) -> int:
        digest = joblib.hash((self._nom, self._err), hash_name="sha1")
        return int.from_bytes(bytes(digest, encoding="utf-8"), "big")

    @property
    def value(self):
        return self._nom

    @property
    def error(self):
        return self._err

    @property
    def relative(self):
        if np.ndim(self._nom) == 0:
            try:
                return self._err / self._nom
            except OverflowError:
                return np.inf
            except ZeroDivisionError:
                return np.NaN
        else:
            return self._err / self._nom

    @property
    def rel(self):
        return self.relative

    @property
    def rel2(self):
        try:
            return self.relative**2
        except OverflowError:
            return np.inf

    def plus_minus(self, err: float):
        val = self._nom
        old_err = self._err
        new_err = np.sqrt(old_err**2 + err**2)

        return self.__class__(val, new_err)

    @classmethod
    def from_string(cls, string: str):
        new_str = string.replace("+/-", "±")
        new_str = new_str.replace("+-", "±")
        if "±" not in new_str:
            return Uncertainty(float(string))
        else:
            u1, u2 = new_str.split("±")
            return cls(float(u1), float(u2))

    @classmethod
    def from_quantities(cls, value, err):
        value_, err_, units = _check_units(value, err)
        inst = cls(value_, err_)
        if units is not None:
            inst *= units
        return inst

    @classmethod
    def from_list(cls, u_list):
        return cls.from_sequence(u_list)

    @classmethod
    def from_sequence(cls, seq):
        _ = iter(seq)

        len_seq = len(seq)
        val = np.empty(len_seq)
        err = np.empty(len_seq)
        if len_seq > 0:
            first_item = seq[0]
            try:
                first_item + 1
            except TypeError:
                raise TypeError(
                    f"Sequence elements of type {type(first_item)} dont support math operations!"
                )
            if hasattr(first_item, "units"):
                val *= first_item.units
                err *= first_item.units
            for i, seq_i in enumerate(seq):
                try:
                    val[i] = float(seq_i._nom)
                    err[i] = float(seq_i._err)
                except AttributeError:
                    val[i] = float(seq_i)
                    err[i] = 0

        return cls(val, err)

    def __float__(self) -> Uncertainty:
        return float(self._nom)

    def __complex__(self) -> Uncertainty:
        return complex(self._nom)

    def __int__(self) -> Uncertainty:
        return int(self._nom)

    # Math Operators
    def __add__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom + other._nom
            new_err = np.sqrt(self._err**2 + other._err**2)
        else:
            new_mag = self._nom + other
            new_err = self._err
        try:
            return self.__class__(new_mag, new_err)
        except NotImplementedError:
            return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom - other._nom
            new_err = np.sqrt(self._err**2 + other._err**2)
        else:
            new_mag = self._nom - other
            new_err = self._err
        try:
            return self.__class__(new_mag, new_err)
        except NotImplementedError:
            return NotImplemented

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom * other._nom
            new_err = np.abs(new_mag) * np.sqrt(self.rel2 + other.rel2)
        else:
            new_mag = self._nom * other
            new_err = np.abs(self._err * other)
        try:
            return self.__class__(new_mag, new_err)
        except NotImplementedError:
            return NotImplemented

    __rmul__ = __mul__

    @ignore_runtime_warnings
    def __truediv__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom / other._nom
            new_err = np.abs(new_mag) * np.sqrt(self.rel2 + other.rel2)
        else:
            new_mag = self._nom / other
            new_err = np.abs(self._err / other)
        try:
            return self.__class__(new_mag, new_err)
        except NotImplementedError:
            return NotImplemented

    @ignore_runtime_warnings
    def __rtruediv__(self, other):
        # Other / Self
        if isinstance(other, Uncertainty):
            raise Exception
        else:
            new_mag = other / self._nom
            new_err = np.abs(new_mag) * np.abs(self.rel)
        try:
            return self.__class__(new_mag, new_err)
        except NotImplementedError:
            return NotImplemented

    __div__ = __truediv__
    __rdiv__ = __rtruediv__

    def __floordiv__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom // other._nom
            new_err = 0.0
        else:
            new_mag = self._nom // other
            new_err = 0.0
        return self.__class__(new_mag, new_err)

    def __rfloordiv__(self, other):
        if isinstance(other, Uncertainty):
            return other.__truediv__(self)
        else:
            new_mag = other // self._nom
            new_err = 0.0
            return self.__class__(new_mag, new_err)

    def __mod__(self, other):
        if isinstance(other, Uncertainty):
            new_mag = self._nom % other._nom
        else:
            new_mag = self._nom % other
        if np.ndim(new_mag) == 0:
            new_err = 0.0
        else:
            new_err = np.zeros_like(new_mag)
        return self.__class__(new_mag, new_err)

    def __rmod__(self, other):
        new_mag = other % self._nom
        if np.ndim(new_mag) == 0:
            new_err = 0.0
        else:
            new_err = np.zeros_like(new_mag)
        return self.__class__(new_mag, new_err)

    def __divmod__(self, other):
        return self // other, self % other

    def __rdivmod__(self, other):
        return other // self, other % self

    @ignore_runtime_warnings
    def __pow__(self, other):
        # Self ** other
        A = self._nom
        sA = self._err
        if isinstance(other, Uncertainty):
            B = other._nom
        else:
            B = other
        new_mag = A**B
        new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2)

        return self.__class__(new_mag, new_err)

    @ignore_runtime_warnings
    def __rpow__(self, other):
        # Other ** self
        B = self._nom
        sB = self._err
        if isinstance(other, Uncertainty):
            A = other._nom
            sA = other._err
        else:
            A = other
            sA = 0

        new_mag = A**B
        new_err = np.abs(new_mag) * np.sqrt(
            (B / A * sA) ** 2 + (np.log(A) * sB) ** 2
        )

        return self.__class__(new_mag, new_err)

    def __abs__(self):
        return self.__class__(abs(self._nom), self._err)

    def __round__(self, ndigits):
        return self.__class__(round(self._nom, ndigits=ndigits), self._err)

    def __pos__(self):
        return self.__class__(operator.pos(self._nom), self._err)

    def __neg__(self):
        return self.__class__(operator.neg(self._nom), self._err)

    def __eq__(self, other):
        if isinstance(other, Uncertainty):
            return self._nom == other._nom
        else:
            return self._nom == other

    def __ne__(self, other):
        out = self.__eq__(other)
        if is_np_duck_array(type(out)):
            return np.logical_not(out)
        else:
            return not out

    def compare(self, other, op):
        if isinstance(other, Uncertainty):
            return op(self._nom, other._nom)
        else:
            return op(self._nom, other)

    __lt__ = lambda self, other: self.compare(  # noqa: E731
        other, op=operator.lt
    )
    __le__ = lambda self, other: self.compare(  # noqa: E731
        other, op=operator.le
    )
    __ge__ = lambda self, other: self.compare(  # noqa: E731
        other, op=operator.ge
    )
    __gt__ = lambda self, other: self.compare(  # noqa: E731
        other, op=operator.gt
    )

    def __bool__(self) -> bool:
        return bool(self._nom)

    __nonzero__ = __bool__

    # NumPy function/ufunc support
    @ignore_runtime_warnings
    def __array_function__(self, func, types, args, kwargs):
        if func.__name__ not in HANDLED_FUNCTIONS:
            return NotImplemented
        elif not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        else:
            return wrap_numpy("function", func, args, kwargs)

    @ignore_runtime_warnings
    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        if method != "__call__":
            raise NotImplementedError
        else:
            if ufunc.__name__ not in HANDLED_UFUNCS:
                raise NotImplementedError(
                    f"Ufunc {ufunc.__name__} is not implemented!"
                ) from None
            else:
                return wrap_numpy("ufunc", ufunc, args, kwargs)

    def __getattr__(self, item):
        if item.startswith("__array_"):
            # Handle array protocol attributes other than `__array__`
            raise AttributeError(
                f"Array protocol attribute {item} not available."
            )
        elif item in self.__apply_to_both_ndarray__:
            val = getattr(self._nom, item)
            err = getattr(self._err, item)

            if callable(val):
                return lambda *args, **kwargs: self.__class__(
                    val(*args, **kwargs), err(*args, **kwargs)
                )
            else:
                return self.__class__(val, err)
        elif item in HANDLED_UFUNCS:
            return lambda *args, **kwargs: wrap_numpy(
                "ufunc", item, [self] + list(args), kwargs
            )
        elif item in HANDLED_FUNCTIONS:
            return lambda *args, **kwargs: wrap_numpy(
                "function", item, [self] + list(args), kwargs
            )
        elif item in self.__ndarray_attributes__:
            return getattr(self._nom, item)
        else:
            raise AttributeError(
                f"Attribute {item} not available in Uncertainty, or as NumPy ufunc or function."
            ) from None

    def __array__(self, t=None) -> np.ndarray:
        warnings.warn(
            "The uncertainty is stripped when downcasting to ndarray.",
            NumpyDowncastWarning,
            stacklevel=2,
        )
        return np.asarray(self._nom)

    def clip(self, min=None, max=None, out=None, **kwargs):
        return self.__class__(
            self._nom.clip(min, max, out, **kwargs), self._err
        )

    def fill(self, value) -> None:
        return self._nom.fill(value)

    def put(self, indices, values, mode="raise") -> None:
        if isinstance(values, self.__class__):
            self._nom.put(indices, values._nom, mode)
            self._err.put(indices, values._err, mode)
        else:
            raise ValueError(
                "Can only 'put' Uncertainties into uncertainties!"
            )

    def copy(self):
        return Uncertainty(self._nom.copy(), self._err.copy())

    # Special properties
    @property
    def flat(self):
        for u, v in (self._nom.flat, self._err.flat):
            yield self.__class__(u, v)

    @property
    def shape(self):
        return self._nom.shape

    @shape.setter
    def shape(self, value):
        self._nom.shape = value
        self._err.shape = value

    @property
    def nbytes(self):
        return self._nom.nbytes + self._err.nbytes

    def searchsorted(self, v, side="left", sorter=None):
        return self._nom.searchsorted(v, side)

    def __len__(self) -> int:
        return len(self._nom)

    def __getitem__(self, key):
        try:
            return self.__class__(self._nom[key], self._err[key])
        except TypeError:
            raise TypeError(f"Index {key} not supported!")

    def __setitem__(self, key, value):
        if not isinstance(value, self.__class__):
            raise ValueError(
                f"Can only pass Uncertainty type to __setitem__! Instead passed {type(value)}"
            )
        try:
            _ = self._nom[key]
        except ValueError as exc:
            raise ValueError(
                f"Object {type(self._nom)} does not support indexing"
            ) from exc

        self._nom[key] = value._nom
        self._err[key] = value._err

    def tolist(self):
        try:
            nom = self._nom.tolist()
            err = self._err.tolist()
            if not isinstance(nom, list):
                return self.__class__(nom, err)
            else:
                return [
                    self.__class__(n, e).tolist()
                    if isinstance(n, list)
                    else self.__class__(n, e)
                    for n, e in (nom, err)
                ]
        except AttributeError:
            raise AttributeError(
                f"{type(self._nom).__name__}' does not support tolist."
            )

    @property
    def ndim(self):
        return np.ndim(self._nom)
